/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <dirent.h> 
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>

/* 1. Type Definition                           */
/************************************************/
typedef signed char         int_08t;
typedef signed short        int_16t;
typedef signed int          int_32t;
typedef signed long long    int_64t;
typedef unsigned char      uint_08t;
typedef unsigned short     uint_16t;
typedef unsigned int       uint_32t;
typedef unsigned long long uint_64t;

typedef uint_32t INST_TYPE;
typedef uint_64t ADDR_TYPE;
typedef uint_64t DATA_TYPE;
typedef uint_64t data_t;


/* 2. Constant Definition                       */
/************************************************/
#define PROJ "SimCore"
#define NAME "Alpha Functional Simulator"
#define VER  "Version 2.0r1 2004-05-11"

#define DEF_FILE_NAME    "aout.txt"
#define BLOCK_SIZE       0x00002000 /* 8KB page */
#define BLOCK_TABLE_SIZE 0x00080000 /* page size*/
#define BLOCK_MASK       0x00001fff /* page mask*/
#define BLOCK_MASK_BIT   13         /* mask bit */
#define DATA_T_SIZE      8   /*  size of data_t */ 
#define DATA_T_BIT       3   /* s bit of data_t */

/** for AMASK op  and IMPLVER op **/
#define IMPLEMENTATION_VERSION 2   
#define CPU_FEATURE_MASK 0x0000000000000107ull 

#define FD_MAPPING_MAX 128      /* read/write  */
#define OSF_SIGCONTEXT_SIZE 568 /* in byte     */
#define OSF_FSTAT_SIZE       80 /* in byte     */
#define OSF_RUSAGE_SIZE      72 /* in byte     */
#define OPTION_MAX          30  /* max options */

/** for systemcall implementation              **/
/** I implement one used for SPEC CPU2000.     **/
/************************************************/
#define ASYS_exit                1
#define ASYS_read                3
#define ASYS_write               4
#define ASYS_close               6
#define ASYS_unlink             10
#define ASYS_brk                17
#define	ASYS_lseek	        19
#define ASYS_getpid             20
#define ASYS_getuid             24
#define ASYS_open               45
#define ASYS_getgid             47
#define ASYS_sigprocmask        48
#define ASYS_ioctl              54
#define ASYS_getpagesize        64
#define ASYS_stat               67
#define ASYS_mmap               71
#define ASYS_munmap             73
#define ASYS_fstat              91
#define ASYS_fcntl              92
#define	ASYS_sigreturn         103
#define	ASYS_gettimeofday      116
#define	ASYS_getrusage         117
#define	ASYS_rename            128
#define	ASYS_getrlimit         144
#define	ASYS_setrlimit         145
#define	ASYS_sigaction         156
#define	ASYS_getdirentries     159
#define ASYS_UNIMPLEMENT       226
#define ASYS_osf_getsysinfo    256
#define ASYS_osf_setsysinfo    257

/** Table C-6: Opcode Summary                  **/
/** by Alpha Architecture Handbook V.4         **/
/************************************************/
/** 0x00  PALcode instruction(CALL_PAL)        **/
/** 0x08  load address instruction             **/
/** 0x09  load address high instruction        **/
/** 0x10  Integer arithmetic instruction       **/
/** 0x11  Integer logical instruction          **/
/** 0x12  Integer shift instruction            **/
/** 0x13  Integer multiply instruction         **/
/** 0x14  Integer to floating-point register   **/
/** 0x15  VAX floating-point instruction       **/
/** 0x16  IEEE floating-point instruction      **/
/** 0x17  Floating-point instruction           **/
/** 0x18  Miscellaneous instruction            **/
/** 0x1a  Jump instruction                     **/
/** 0x1c  Floating-point to integer register   **/
/************************************************/
#define OP_PAL  0x00
#define OP_LDA  0x08
#define OP_LDAH 0x09
#define OP_INTA 0x10
#define OP_INTL 0x11
#define OP_INTS 0x12
#define OP_INTM 0x13
#define OP_ITFP 0x14
#define OP_FLTV 0x15
#define OP_FLTI 0x16
#define OP_FLTL 0x17
#define OP_MISC 0x18
#define OP_JSR  0x1a
#define OP_FPTI 0x1c

#define BIT0    0x0000000000000001ull
#define BIT1    0x0000000000000002ull
#define BIT2    0x0000000000000004ull
#define BIT3    0x0000000000000008ull
#define BIT4    0x0000000000000010ull
#define BIT5    0x0000000000000020ull
#define BIT6    0x0000000000000040ull
#define BIT7    0x0000000000000080ull
#define BIT8    0x0000000000000100ull
#define BIT9    0x0000000000000200ull
#define BIT10   0x0000000000000400ull
#define BIT11   0x0000000000000800ull
#define BIT12   0x0000000000001000ull
#define BIT15   0x0000000000008000ull
#define BIT20   0x0000000000100000ull
#define BIT31   0x0000000080000000ull
#define BIT63   0x8000000000000000ull

#define EXTND8  0xffffffffffffff00ull
#define EXTND16 0xffffffffffff0000ull
#define EXTND21 0xffffffffffe00000ull
#define EXTND32 0xffffffff00000000ull
#define MASK08  0x00000000000000ffull
#define MASK16  0x000000000000ffffull
#define MASK21  0x00000000001fffffull
#define MASK32  0x00000000ffffffffull

#define ZAP0    0xffffffffffffff00ull
#define ZAP1    0xffffffffffff00ffull
#define ZAP2    0xffffffffff00ffffull
#define ZAP3    0xffffffff00ffffffull
#define ZAP4    0xffffff00ffffffffull
#define ZAP5    0xffff00ffffffffffull
#define ZAP6    0xff00ffffffffffffull
#define ZAP7    0x00ffffffffffffffull

/** for instruction.cc **/
#define MSK1 0x3e  /* 6bit of 111110 */
#define MSK2 0x3c  /* 6bit of 111100 */
#define MSK3 0x38  /* 6bit of 111000 */
#define MSK4 0x30  /* 6bit of 110000 */

/*** Control Transfer instructions            ***/
/************************************************/
#define _JMP__   0x1a0000 // 1a.0
#define _JSR__   0x1a0001 // 1a.1
#define _RET__   0x1a0002 // 1a.2
#define _JSR_C   0x1a0003 // 1a.3 JSR_COROUTINE

#define _BR__    0x300000
#define _FBEQ    0x310000
#define _FBLT    0x320000
#define _FBLE    0x330000
#define _BSR_    0x340000 
#define _FBNE    0x350000 
#define _FBGE    0x360000 
#define _FBGT    0x370000
#define _BLGC    0x380000 
#define _BEQ_    0x390000 
#define _BLT_    0x3a0000 
#define _BLE_    0x3b0000 
#define _BLBS    0x3c0000 
#define _BNE_    0x3d0000 
#define _BGE_    0x3e0000
#define _BGT_    0x3f0000

/*** Memory Access & LDA & LDAH Instructions  ***/
/************************************************/
#define _LDA___  0x080000
#define _LDAH__  0x090000
#define _OPC0A_  0x0a0000
#define _LDQ_U_  0x0b0000
#define _OPC0C_  0x0c0000
#define _OPC0D_  0x0d0000
#define _OPC0E_  0x0e0000
#define _STQ_U_  0x0f0000

#define _LDF___  0x200000
#define _LDG___  0x210000
#define _LDS___  0x220000
#define _LDT___  0x230000
#define _STF___  0x240000
#define _STG___  0x250000
#define _STS___  0x260000
#define _STT___  0x270000
#define _LDL___  0x280000
#define _LDQ___  0x290000
#define _LDL_L_  0x2a0000
#define _LDQ_L_  0x2b0000
#define _STL___  0x2c0000
#define _STQ___  0x2d0000
#define _STL_C_  0x2e0000
#define _STQ_C_  0x2f0000

/*** Arithmetic instructions (int and float)  ***/
/************************************************/
#define _ADDL__  0x100000
#define _S4ADDL  0x100002
#define _SUBL__  0x100009
#define _S4SUBL  0x10000b
#define _CMPBGE  0x10000f
#define _S8ADDL  0x100012
#define _S8SUBL  0x10001b
#define _CMPULT  0x10001d
#define _ADDQ__  0x100020
#define _S4ADDQ  0x100022
#define _SUBQ__  0x100029
#define _S4SUBQ  0x10002b
#define _CMPEQ_  0x10002d
#define _S8ADDQ  0x100032
#define _S8SUBQ  0x10003b
#define _CMPULE  0x10003d
#define _CMPLT_  0x10004d
#define _CMPLE_  0x10006d

/*** Arithmetic Logical   ***/
/************************************************/
#define _AND___  0x110000
#define _BIC___  0x110008
#define _CMOVLBS 0x110014
#define _CMOVLBC 0x110016
#define _BIS___  0x110020
#define _CMOVEQ  0x110024
#define _CMOVNE  0x110026
#define _ORNOT_  0x110028
#define _XOR___  0x110040
#define _CMOVLT  0x110044
#define _CMOVGE  0x110046
#define _EQV___  0x110048
#define _AMASK_  0x110061
#define _CMOVLE  0x110064
#define _CMOVGT  0x110066
#define _IMPLVER 0x11006c

/*** Integer shift instruction                ***/
/************************************************/
#define _SRL___  0x120034
#define _SLL___  0x120039
#define _SRA___  0x12003c
#define _ZAP___  0x120030
#define _ZAPNOT  0x120031
#define _EXTBL_  0x120006
#define _EXTWL_  0x120016
#define _EXTLL_  0x120026
#define _EXTQL_  0x120036
#define _EXTWH_  0x12005a
#define _EXTLH_  0x12006a
#define _EXTQH_  0x12007a
#define _INSBL_  0x12000b
#define _INSWL_  0x12001b
#define _INSLL_  0x12002b
#define _INSQL_  0x12003b
#define _MSKBL_  0x120002
#define _MSKWL_  0x120012
#define _MSKLL_  0x120022
#define _MSKQL_  0x120032
#define _MSKWH_  0x120052
#define _MSKLH_  0x120062
#define _MSKQH_  0x120072
#define _INSWH_  0x120057
#define _INSLH_  0x120067
#define _INSQH_  0x120077

/*** Floating instruction (a part of)         ***/
/************************************************/
#define _FCMOVEQ 0x17002a
#define _FCMOVNE 0x17002b
#define _FCMOVLT 0x17002c
#define _FCMOVGE 0x17002d
#define _FCMOVLE 0x17002e
#define _FCMOVGT 0x17002f

/*** Multiply instruction                     ***/
/************************************************/
#define _MULL__  0x130000
#define _MULQ__  0x130020
#define _UMULH_  0x130030

/************************************************/


/* 4. structure and class definition            */
/************************************************/
/*** Used in syscall.c SYS_fstat ***/
struct osf_statbuf{ 
  uint_32t a_st_dev;
  uint_32t a_st_ino;
  uint_32t a_st_mode;
  uint_16t a_st_nlink;
  uint_16t a_pad0;
  uint_32t a_st_uid;
  uint_32t a_st_gid;
  uint_32t a_st_rdev;
  uint_32t a_pad1;
  uint_64t a_st_size;
  uint_32t a_st_atime;
  uint_32t a_st_spare1;
  uint_32t a_st_mtime;
  uint_32t a_st_spare2;
  uint_32t a_st_ctime;
  uint_32t a_st_spare3;
  uint_32t a_st_blksize;
  uint_32t a_st_blocks;
  uint_32t a_st_gennum;
  uint_32t a_st_spare4;
};

/************************************************/
class Env{
 public:
  class system_config      *sc;
  class evaluation_result  *e;
  class debug              *deb;
  class system_manager     *sys;
  class memory_system      *mem;
  class architecture_state *as;
};

/************************************************/
class system_config{
 public:
  char program_name[512];
  uint_64t end_tc;
  uint_64t trace_flag;
  int slow_mode;
  int verbose_interval;
  int debug_flag;
  int distrib_flag;
  int verify_flag;
  int syscall_flag;
  int result_flag;
  int simple_result;
  int fd_adjust;
  int fd_stdin_flag;
  int fd_stdin; /* file descripter for STDIN */
  system_config(char *, char**);
};

/************************************************/
class main_memory{
  data_t *block_table[BLOCK_TABLE_SIZE];
  data_t *allocblock(data_t *);
 public:
  main_memory();
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
  int load_all(data_t*, data_t*);
};

/************************************************/
class evaluation_result{
 public:
  uint_64t retired_inst; 
  int used_memory_block; 
  time_t time_begin;     /* start time stamp */
  struct timeval tp;     /* start time stamp */
  struct timezone tzp;   /* start time stamp */
  evaluation_result();
  main_memory *imix;    /* for instructio mix */
};

/************************************************/
class architecture_state{
  Env *ev;
 public:
  data_t pc;        /* program counter      */
  data_t r[32];     /* general purpose regs */
  data_t f[32];     /* floating point  regs */
  uint_64t rr(int); /* register read        */
  architecture_state(Env*);
  architecture_state();
};

/************************************************/
class memory_system{
  Env *ev;
 public:
  class main_memory *mm;
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
  void ld_inst(data_t *, INST_TYPE *);
  void ld_nbyte(int, data_t *, data_t *);
  void st_nbyte(int, data_t *, data_t *);
  ~memory_system();
  memory_system(Env*);
};

/************************************************/
class system_manager{
  Env *ev;
  /* File Descriptor Mapping Table */
  DATA_TYPE fd_table[FD_MAPPING_MAX]; 
  /* Unique Data for rduniq & wruniq inst */
  DATA_TYPE uniq; 
 public:
  int running;    /* chip is running ? */
  void execute_pal(class instruction*);
  int translate_fd(DATA_TYPE);
  system_manager(Env*);
};

/************************************************/
class debug{
  Env *ev;
 public:
  debug(Env*);
  void debugmode(instruction*);
};

/************************************************/
class instruction{
 public:
  Env *ev;      /* pointer to the environment */
  INST_TYPE ir; /* 32b: instruction code      */
  int  CI;      /* 24b: Code ID               */
  char Op;      /*  6b: Opcode field          */
  char BR;      /*  1b: branch inst ?         */
  char CM;      /*  1b: CMOV inst ?           */
  char ST;      /*  4b: store inst & byte num */
  char LD;      /*  4b: load inst  & byte num */
  char Ai;      /*  1b: Rav is immediate?     */
  char Bi;      /*  1b: Rbv is immediate?     */
  char Ar;      /*  1b: Rav from general-reg? */
  char Br;      /*  1b: Rbv from general-reg? */
  char RA;      /*  5b: Rav reg-read  index   */
  char RB;      /*  5b: Rbv reg-read  index   */
  char WB;      /*  5b: Writeback reg index   */
  data_t  Cpc;  /* PC of the instruction      */
  data_t  Npc;  /* PC of the next inst.       */
                /* Calculated in Exe. stage   */
  data_t  Imm;  /* immediate                  */
  data_t  Adr;  /* load & store : mem address */
                /* cmov inst.   : Rmv value   */
  data_t  Rav;  /* Ra                         */
  data_t  Rbv;  /* Rb                         */
  data_t  Rcv;  /* Rc                         */
  void Fetch(data_t*); /* stage 0 */
  void Slot();         /* stage 1 */
  void Rename();       /* stage 2 */
  void Issue();        /* stage 3 */
  void RegisterRead(); /* stage 4 */
  void Execute();      /* stage 5 */
  void Memory();       /* stage 6 */
  void WriteBack();    /* stage 7 */
  void BackEnd();      /* stage 4,5,6,7 */
  instruction(Env*);
};

/************************************************/
class simple_chip{
 public:
  Env         *ev;
  instruction *p;
  simple_chip(char*, char**);
  ~simple_chip();
  int step();
  void loop();
};

/************************************************/
struct alpha_stat{  /*** for syscall.cc ***/
  uint_64t  a_st_dev;         
  uint_32t  a_st_ino;              
   int_32t  a___pad1;
  uint_32t  a_st_mode;            
  uint_32t  a_st_nlink;          
  uint_32t  a_st_uid;              
  uint_32t  a_st_gid;              
  uint_64t  a_st_rdev;             
   int_64t  a_st_size;             
   int_64t  a_st_atime;           
   int_64t  a_st_mtime;           
   int_64t  a_st_ctime;
  uint_32t  a_st_blocks;        
   int_32t  a___pad2;
  uint_32t  a_st_blksize;     
  uint_32t  a_st_flags;
  uint_32t  a_st_gen;
   int_32t  a___pad3;
   int_64t  a___unused[4];
};


/* 5. function prototypes                       */
/************************************************/
extern void print_evaluation_result(Env*);
extern void usage();
extern uint_64t map_s(DATA_TYPE);
extern int cmov_ir(INST_TYPE);
extern void option_from_file(char**, char*);
extern void house_keeper(instruction*);
extern "C" int simple_cpu(int, uint_64t*);
extern uint_64t get_hex(char*);
extern int get_code(INST_TYPE);
extern int ld_bytes(int);
extern int st_bytes(int);
extern DATA_TYPE sext(DATA_TYPE);
extern DATA_TYPE sext16(DATA_TYPE);
extern DATA_TYPE sext21(DATA_TYPE);
extern int WR(int);
extern void set_ld_data(int, int, data_t*, 
			data_t*, data_t*);
extern void set_st_data(int, int, data_t*, 
			data_t*, data_t*, 
			DATA_TYPE*);
extern void UnitA(int, DATA_TYPE, 
		  DATA_TYPE, DATA_TYPE*);
extern void UnitL(int, DATA_TYPE, 
		  DATA_TYPE, DATA_TYPE*);
extern void UnitS(int, DATA_TYPE, 
		  DATA_TYPE, DATA_TYPE*);
extern void UnitM(int, DATA_TYPE, 
		  DATA_TYPE, DATA_TYPE*);
extern void FuITFP(INST_TYPE, DATA_TYPE, 
		   DATA_TYPE, DATA_TYPE*);
extern void FuFLTL(INST_TYPE, DATA_TYPE, 
		   DATA_TYPE, DATA_TYPE*);
extern void FuFPTI(INST_TYPE, DATA_TYPE, 
		   DATA_TYPE, DATA_TYPE*);
extern void FuFLTI(INST_TYPE, DATA_TYPE, 
		   DATA_TYPE, DATA_TYPE*);
extern void CmovUnit(int, DATA_TYPE, int*);
extern int Btaken(int, DATA_TYPE);
extern void print_register_name(int);
/************************************************/
